/*
 * arch/arm/mach-ep93xx/ts72xx.c
 * Technologic Systems TS72xx SBC support.
 *
 * Copyright (C) 2006 Lennert Buytenhek <buytenh@wantstofly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/mtd/physmap.h>
#include <linux/platform_device.h>
#include <linux/m48t86.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/hardware.h>
#include <asm/mach-types.h>
#include <asm/mach/arch.h>
#include <asm/mach/map.h>

volatile static unsigned char *ts7350_rtc = NULL;
volatile static unsigned char *ts7390_rtc = NULL;
volatile static unsigned short *ts7350_pc104 = NULL;
static int config_ts7390_rtc = 0;
static int config_ts7400_rtc = 0;

//TS7400 RTC GPIO regs
volatile unsigned int *dat_dr, *dat_ddr, *cs_dr, *cs_ddr, *ale_ddr, *ale_dr;
volatile unsigned int *rd_dr, *rd_ddr, *wr_dr, *wr_ddr;
#define DAT_DR		0x40000
#define DAT_DDR		0x40010
#define CS_DR		0x40040
#define CS_DDR		0x40044
#define ALE_DR		0x40040
#define ALE_DDR		0x40044
#define RD_DR		0x40040
#define RD_DDR		0x40044
#define WR_DR		0x40030
#define WR_DDR		0x40034

static struct map_desc ts7xxx_io_desc[] __initdata = {
{
	 .virtual = TS72XX_MODEL_VIRT_BASE,
	 .pfn = __phys_to_pfn(TS72XX_MODEL_PHYS_BASE),
	 .length = TS72XX_MODEL_SIZE,
	 .type = MT_DEVICE,
	 }, {
	     .virtual = TS72XX_OPTIONS_VIRT_BASE,
	     .pfn = __phys_to_pfn(TS72XX_OPTIONS_PHYS_BASE),
	     .length = TS72XX_OPTIONS_SIZE,
	     .type = MT_DEVICE,
	     }, {
		 .virtual = TS72XX_OPTIONS2_VIRT_BASE,
		 .pfn = __phys_to_pfn(TS72XX_OPTIONS2_PHYS_BASE),
		 .length = TS72XX_OPTIONS2_SIZE,
		 .type = MT_DEVICE,
		 }, {
		     .virtual = TS72XX_RTC_INDEX_VIRT_BASE,
		     .pfn = __phys_to_pfn(TS72XX_RTC_INDEX_PHYS_BASE),
		     .length = TS72XX_RTC_INDEX_SIZE,
		     .type = MT_DEVICE,
		     }, {
			 .virtual = TS72XX_RTC_DATA_VIRT_BASE,
			 .pfn = __phys_to_pfn(TS72XX_RTC_DATA_PHYS_BASE),
			 .length = TS72XX_RTC_DATA_SIZE,
			 .type = MT_DEVICE,
			 }, {
			     .virtual = TS73XX_MODEL_VIRT_BASE,
			     .pfn = __phys_to_pfn(TS73XX_MODEL_PHYS_BASE),
			     .length = TS73XX_MODEL_SIZE,
			     .type = MT_DEVICE,
			     }
};

static struct map_desc ts72xx_nand_io_desc[] __initdata = {
	{
	 .virtual = TS72XX_NAND_DATA_VIRT_BASE,
	 .pfn = __phys_to_pfn(TS72XX_NAND1_DATA_PHYS_BASE),
	 .length = TS72XX_NAND_DATA_SIZE,
	 .type = MT_DEVICE,
	 }, {
	     .virtual = TS72XX_NAND_CONTROL_VIRT_BASE,
	     .pfn = __phys_to_pfn(TS72XX_NAND1_CONTROL_PHYS_BASE),
	     .length = TS72XX_NAND_CONTROL_SIZE,
	     .type = MT_DEVICE,
	     }, {
		 .virtual = TS72XX_NAND_BUSY_VIRT_BASE,
		 .pfn = __phys_to_pfn(TS72XX_NAND1_BUSY_PHYS_BASE),
		 .length = TS72XX_NAND_BUSY_SIZE,
		 .type = MT_DEVICE,
		 }
};

static struct map_desc ts72xx_alternate_nand_io_desc[] __initdata = {
	{
	 .virtual = TS72XX_NAND_DATA_VIRT_BASE,
	 .pfn = __phys_to_pfn(TS72XX_NAND2_DATA_PHYS_BASE),
	 .length = TS72XX_NAND_DATA_SIZE,
	 .type = MT_DEVICE,
	 }, {
	     .virtual = TS72XX_NAND_CONTROL_VIRT_BASE,
	     .pfn = __phys_to_pfn(TS72XX_NAND2_CONTROL_PHYS_BASE),
	     .length = TS72XX_NAND_CONTROL_SIZE,
	     .type = MT_DEVICE,
	     }, {
		 .virtual = TS72XX_NAND_BUSY_VIRT_BASE,
		 .pfn = __phys_to_pfn(TS72XX_NAND2_BUSY_PHYS_BASE),
		 .length = TS72XX_NAND_BUSY_SIZE,
		 .type = MT_DEVICE,
		 }
};

static void __init ts72xx_map_io(void)
{
	ep93xx_map_io();
	iotable_init(ts7xxx_io_desc, ARRAY_SIZE(ts7xxx_io_desc));

	/*
	 * The TS-7200 has NOR flash, the other models have NAND flash.
	 */
	if (!board_is_ts7200()) {
		if (is_ts9420_installed()) {
			iotable_init(ts72xx_alternate_nand_io_desc,
				     ARRAY_SIZE(ts72xx_alternate_nand_io_desc));
		} else {
			iotable_init(ts72xx_nand_io_desc,
				     ARRAY_SIZE(ts72xx_nand_io_desc));
		}
	}
}

static struct physmap_flash_data ts72xx_flash_data = {
	.width = 1,
};

static struct resource ts72xx_flash_resource = {
	.start = TS72XX_NOR_PHYS_BASE,
	.end = TS72XX_NOR_PHYS_BASE + 0x00ffffff,
	.flags = IORESOURCE_MEM,
};

static struct platform_device ts72xx_flash = {
	.name = "physmap-flash",
	.id = 0,
	.dev = {
		.platform_data = &ts72xx_flash_data,
		},
	.num_resources = 1,
	.resource = &ts72xx_flash_resource,
};

static void inline setRTCdata(unsigned char ch)
{
	*dat_dr = (unsigned int)ch;
	udelay(10);
	*dat_ddr = 0xFF;
}

static unsigned char inline getRTCdata(void)
{
	*dat_ddr = 0x00;
	udelay(10);
	return (unsigned char)(*dat_dr & 0xFF);
}

static void inline setRTC_RD(int en)
{
	if (!en)
		*rd_dr |= (1 << 3);
	else
		*rd_dr &= ~(1 << 3);
	*rd_ddr |= (1 << 3);
}

static void inline setRTC_CS(int en)
{
	if (!en)
		*cs_dr |= (1 << 5);
	else
		*cs_dr &= ~(1 << 5);
	*cs_ddr |= (1 << 5);
	udelay(10);
}

static void inline setRTC_ALE(int en)
{
	if (en)
		*ale_dr |= (1 << 4);
	else
		*ale_dr &= ~(1 << 4);
	*ale_ddr |= (1 << 4);
	udelay(10);
}

static void inline setRTC_WR(int en)
{
	if (!en)
		*wr_dr |= (1 << 3);
	else
		*wr_dr &= ~(1 << 3);
	*wr_ddr |= (1 << 3);
	udelay(10);
}

static void ts7400_rtc_writebyte(unsigned char value, unsigned long addr)
{
	if (!config_ts7400_rtc) {
		ale_ddr = (unsigned int *)(EP93XX_APB_VIRT_BASE + ALE_DDR);
		ale_dr = (unsigned int *)(EP93XX_APB_VIRT_BASE + ALE_DR);
		cs_ddr = (unsigned int *)(EP93XX_APB_VIRT_BASE + CS_DDR);
		cs_dr = (unsigned int *)(EP93XX_APB_VIRT_BASE + CS_DR);
		wr_ddr = (unsigned int *)(EP93XX_APB_VIRT_BASE + WR_DDR);
		wr_dr = (unsigned int *)(EP93XX_APB_VIRT_BASE + WR_DR);
		rd_ddr = (unsigned int *)(EP93XX_APB_VIRT_BASE + RD_DDR);
		rd_dr = (unsigned int *)(EP93XX_APB_VIRT_BASE + RD_DR);
		dat_ddr = (unsigned int *)(EP93XX_APB_VIRT_BASE + DAT_DDR);
		dat_dr = (unsigned int *)(EP93XX_APB_VIRT_BASE + DAT_DR);

		__raw_writel(0xaa, EP93XX_SYSCON_SWLOCK);
		__raw_writel(0x08140d00, EP93XX_SYSCON_DEVICE_CONFIG);

		config_ts7400_rtc = 1;
		//ensure RTC battery is enabled
		ts7400_rtc_writebyte(0x20, 0xa);	//enable the battery
	}
	setRTC_ALE(0);
	setRTC_RD(0);
	setRTC_CS(0);
	setRTC_WR(0);
	setRTC_RD(0);
	setRTC_CS(1);
	setRTC_ALE(1);
	setRTCdata(addr);
	setRTC_ALE(0);
	setRTC_WR(1);
	setRTCdata(value);
	setRTC_WR(0);
	setRTC_CS(0);
}

static unsigned char ts7400_rtc_readbyte(unsigned long addr)
{
	unsigned char dat;
	if (!config_ts7400_rtc) {
		ale_ddr = (unsigned int *)(EP93XX_APB_VIRT_BASE + ALE_DDR);
		ale_dr = (unsigned int *)(EP93XX_APB_VIRT_BASE + ALE_DR);
		cs_ddr = (unsigned int *)(EP93XX_APB_VIRT_BASE + CS_DDR);
		cs_dr = (unsigned int *)(EP93XX_APB_VIRT_BASE + CS_DR);
		wr_ddr = (unsigned int *)(EP93XX_APB_VIRT_BASE + WR_DDR);
		wr_dr = (unsigned int *)(EP93XX_APB_VIRT_BASE + WR_DR);
		rd_ddr = (unsigned int *)(EP93XX_APB_VIRT_BASE + RD_DDR);
		rd_dr = (unsigned int *)(EP93XX_APB_VIRT_BASE + RD_DR);
		dat_ddr = (unsigned int *)(EP93XX_APB_VIRT_BASE + DAT_DDR);
		dat_dr = (unsigned int *)(EP93XX_APB_VIRT_BASE + DAT_DR);

		__raw_writel(0xaa, EP93XX_SYSCON_SWLOCK);
		__raw_writel(0x08140d00, EP93XX_SYSCON_DEVICE_CONFIG);

		config_ts7400_rtc = 1;
		//ensure RTC battery is enabled
		ts7400_rtc_writebyte(0x20, 0xa);	//enable the battery
	}
	setRTC_ALE(0);
	setRTC_RD(0);
	setRTC_CS(0);
	setRTC_WR(0);
	setRTC_RD(0);
	setRTC_CS(1);
	setRTC_ALE(1);
	setRTCdata(addr);
	setRTC_ALE(0);
	getRTCdata();
	setRTC_RD(1);
	dat = getRTCdata();
	setRTC_RD(0);
	setRTC_CS(0);
	return dat;
}

static unsigned char ts72xx_rtc_readbyte(unsigned long addr)
{
	__raw_writeb(addr, TS72XX_RTC_INDEX_VIRT_BASE);
	return __raw_readb(TS72XX_RTC_DATA_VIRT_BASE);
}

static void ts72xx_rtc_writebyte(unsigned char value, unsigned long addr)
{
	__raw_writeb(addr, TS72XX_RTC_INDEX_VIRT_BASE);
	__raw_writeb(value, TS72XX_RTC_DATA_VIRT_BASE);
}

#define RESTORE_PC104 config_pc104(1)
#define SETUP_PC104 config_pc104(0)

static inline void config_pc104(int restore)
{
	static unsigned short a_lower, a_upper, b_lower, b_upper, isaconf;

	if (restore) {

		//save PC/104 status regs
		ts7350_pc104[TS7350_PC104_A_LOW] = a_lower;
		ts7350_pc104[TS7350_PC104_A_UPPER] = a_upper;
		ts7350_pc104[TS7350_PC104_B_LOW] = b_lower;
		ts7350_pc104[TS7350_PC104_B_UPPER] = b_upper;
		ts7350_pc104[TS7350_PC104_B_UPPER] = isaconf;

	} else {

		if (!ts7350_rtc)
			ts7350_rtc = (unsigned char *)
			    ioremap(TS7350_RTC_BASE, TS7350_RTC_SIZE);

		if (!ts7350_pc104)
			ts7350_pc104 = (unsigned short *)
			    ioremap(TS7350_PC104_BASE, TS7350_PC104_SIZE);

		//save PC/104 status regs
		a_lower = ts7350_pc104[TS7350_PC104_A_LOW];
		a_upper = ts7350_pc104[TS7350_PC104_A_UPPER];
		b_lower = ts7350_pc104[TS7350_PC104_B_LOW];
		b_upper = ts7350_pc104[TS7350_PC104_B_UPPER];
		isaconf = ts7350_pc104[TS7350_PC104_ISACONF];

		//configure PC/104 connector for PC/104 bus
		//we need to do this to talk to the RTC
		ts7350_pc104[TS7350_PC104_A_LOW] = 0x5555;
		ts7350_pc104[TS7350_PC104_A_UPPER] = 0x5555;
		ts7350_pc104[TS7350_PC104_B_LOW] = 0x5555;
		ts7350_pc104[TS7350_PC104_B_UPPER] = 0x5555;

		//setup PC/104 setup time and hold time
		ts7350_pc104[TS7350_PC104_ISACONF] = 0x1ef;

		//ensure RTC battery is enabled
		ts7350_rtc[TS7350_RTC_INDEX] = 0xa;
		ts7350_rtc[TS7350_RTC_DATA] = 0x20;
	}
}

static unsigned char ts7350_rtc_readbyte(unsigned long addr)
{
	unsigned char value;

	SETUP_PC104;
	ts7350_rtc[TS7350_RTC_INDEX] = addr;
	value = ts7350_rtc[TS7350_RTC_DATA];
	RESTORE_PC104;

	return value;
}

static void ts7350_rtc_writebyte(unsigned char value, unsigned long addr)
{

	SETUP_PC104;
	ts7350_rtc[TS7350_RTC_INDEX] = addr;
	ts7350_rtc[TS7350_RTC_DATA] = value;
	RESTORE_PC104;
}

static unsigned char ts7390_rtc_readbyte(unsigned long addr)
{
	unsigned char value;

	if (!config_ts7390_rtc) {
		ts7390_rtc = (unsigned char *)
		    ioremap(TS7390_RTC_BASE, TS7390_RTC_SIZE);
		//ensure RTC battery is enabled
		ts7390_rtc[TS7390_RTC_INDEX] = 0xa;
		ts7390_rtc[TS7390_RTC_DATA] = 0x20;
		config_ts7390_rtc = 1;
	}

	ts7390_rtc[TS7390_RTC_INDEX] = addr;
	value = ts7390_rtc[TS7390_RTC_DATA];

	return value;
}

static void ts7390_rtc_writebyte(unsigned char value, unsigned long addr)
{
	if (!config_ts7390_rtc) {
		ts7390_rtc = (unsigned char *)
		    ioremap(TS7390_RTC_BASE, TS7390_RTC_SIZE);
		//ensure RTC battery is enabled
		ts7390_rtc[TS7390_RTC_INDEX] = 0xa;
		ts7390_rtc[TS7390_RTC_DATA] = 0x20;
		config_ts7390_rtc = 1;
	}
	ts7390_rtc[TS7390_RTC_INDEX] = addr;
	ts7390_rtc[TS7390_RTC_DATA] = value;
}

static struct m48t86_ops ts72xx_rtc_ops = {

	.readbyte = ts72xx_rtc_readbyte,
	.writebyte = ts72xx_rtc_writebyte,
};

static struct m48t86_ops ts7350_rtc_ops = {

	.readbyte = ts7350_rtc_readbyte,
	.writebyte = ts7350_rtc_writebyte,
};

static struct m48t86_ops ts7390_rtc_ops = {

	.readbyte = ts7390_rtc_readbyte,
	.writebyte = ts7390_rtc_writebyte,
};

static struct platform_device ts72xx_rtc_device = {
	.name = "rtc-m48t86",
	.id = -1,
	.dev = {
		.platform_data = &ts72xx_rtc_ops,
		},
	.num_resources = 0,
};

static struct platform_device ts7350_rtc_device = {
	.name = "rtc-m48t86",
	.id = -1,
	.dev = {
		.platform_data = &ts7350_rtc_ops,
		},
	.num_resources = 0,
};

static struct platform_device ts7390_rtc_device = {
	.name = "rtc-m48t86",
	.id = -1,
	.dev = {
		.platform_data = &ts7390_rtc_ops,
		},
	.num_resources = 0,
};

static struct ep93xx_eth_data ts72xx_eth_data = {
	.phy_id = 1,
};

static struct resource ts72xx_eth_resource[] = {
	{
	 .start = EP93XX_ETHERNET_PHYS_BASE,
	 .end = EP93XX_ETHERNET_PHYS_BASE + 0xffff,
	 .flags = IORESOURCE_MEM,
	 }, {
	     .start = IRQ_EP93XX_ETHERNET,
	     .end = IRQ_EP93XX_ETHERNET,
	     .flags = IORESOURCE_IRQ,
	     }
};

static struct platform_device ts72xx_eth_device = {
	.name = "ep93xx-eth",
	.id = -1,
	.dev = {
		.platform_data = &ts72xx_eth_data,
		},
	.num_resources = 2,
	.resource = ts72xx_eth_resource,
};

static struct resource ts72xx_watchdog_resources[] = {
	[0] = {
	       .start = TS72XX_WATCHDOG_CONTROL_PHYS_BASE,
	       .end = TS72XX_WATCHDOG_CONTROL_PHYS_BASE + 0x0fff,
	       .flags = IORESOURCE_MEM,
	       },
	[1] = {
	       .start = TS72XX_WATCHDOG_FEED_PHYS_BASE,
	       .end = TS72XX_WATCHDOG_FEED_PHYS_BASE + 0x0fff,
	       .flags = IORESOURCE_MEM,
	       },
};

static struct platform_device ts72xx_watchdog_device = {
	.name = "ts72xx_wdt",
	.id = -1,
	.num_resources = ARRAY_SIZE(ts72xx_watchdog_resources),
	.resource = ts72xx_watchdog_resources,
};

static struct m48t86_ops ts7400_rtc_ops = {

	.readbyte = ts7400_rtc_readbyte,
	.writebyte = ts7400_rtc_writebyte,
};

static struct platform_device ts7400_rtc_device = {
	.name = "rtc-m48t86",
	.id = -1,
	.dev = {
		.platform_data = &ts7400_rtc_ops,
		},
	.num_resources = 0,
};

static void __init ts72xx_init_machine(void)
{
	ep93xx_init_devices();

	//EWD: we should consider using .config entries, as these
	//detection routines read id registers which are in different
	//locations on different boards. And even worse the 7200 detection
	//routine will return true for a 7350(e.g. 0x2200_0000 & 0x7 == 0),
	//so we must test for a 73x0 first(e.g. order matters in the test
	//below)!!! If you add a new board_is_xxx below make sure to test 
	//the new kernel on all boards...
	//
#ifdef CONFIG_MACH_PROSOFT
	platform_device_register(&ts7400_rtc_device);
#else
	if (board_is_ts7350() || board_is_ts7370()) {
		platform_device_register(&ts7350_rtc_device);
	} else if (board_is_ts7390()) {
		platform_device_register(&ts7390_rtc_device);
	} else if (board_is_ts7200()) {
		platform_device_register(&ts72xx_flash);
	} else if (board_is_ts7400()) {
		platform_device_register(&ts7400_rtc_device);
	} else {
		platform_device_register(&ts72xx_rtc_device);
	}
#endif

	memcpy(ts72xx_eth_data.dev_addr,
	       (void *)(EP93XX_ETHERNET_BASE + 0x50), 6);
	platform_device_register(&ts72xx_eth_device);
	platform_device_register(&ts72xx_watchdog_device);
}

MACHINE_START(TS72XX, "Technologic Systems TS-72xx SBC")
    /* Maintainer: Lennert Buytenhek <buytenh@wantstofly.org> */
    .phys_io = EP93XX_APB_PHYS_BASE,
    .io_pg_offst = ((EP93XX_APB_VIRT_BASE) >> 18) & 0xfffc,
    .boot_params = 0x00000100,
    .map_io = ts72xx_map_io,
    .init_irq = ep93xx_init_irq,
    .timer = &ep93xx_timer,
    .init_machine = ts72xx_init_machine, 
MACHINE_END
