/*
* EHCI HCD (Host Controller Driver) for USB.
*
* Written by Bob Perry (2007) lb-source@users.sourceforge.net
*
* Bus Glue for Marvel System On Chip 5182
*
* Based on fragments of Marvells driver 
*
* Device mode is not implemented
* Suspend/resume is TBD
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <linux/platform_device.h>
#include <asm/arch-mv88fxx81/usb.h>

// FIXME: Power Managment is un-ported so temporarily disable it
#undef CONFIG_PM

static int mv5182_start_ehc(struct platform_device *pdev)
{
    int winNum;
    unsigned long regVal;
    unsigned int size, base;
	unsigned int baseout[]={
		0x0, 			// System memory
		0xE0000000, 	// PCI memory
		0xE8000000		// PCI memory
		};
	int iDevice=0;
	unsigned long timeout=10000000;
	
	// Get USB port number
	iDevice = pdev->id;
	
	if (iDevice < 0 || iDevice > 1) {
		printk("%s: Does not support more than 2 ports\n", __FILE__);
		return -ENODEV;
	}
		
	printk(KERN_DEBUG "Starting 88F5182 SOC EHCI USB Controller\n");

    // Clear Interrupt Cause and Mask registers
    mv_raw_write_reg32(MV_USB_BRIDGE_INTR_CAUSE_REG(iDevice), 0);
    mv_raw_write_reg32(MV_USB_BRIDGE_INTR_MASK_REG(iDevice), 0);

    // Reset controller
    regVal = mv_raw_read_reg32(MV_USB_CORE_CMD_REG(iDevice));
    mv_raw_write_reg32(MV_USB_CORE_CMD_REG(iDevice), regVal | MV_USB_CORE_CMD_RESET_MASK);
	
	// Wait for controller reset to complete
    while(( mv_raw_read_reg32(MV_USB_CORE_CMD_REG(iDevice)) & MV_USB_CORE_CMD_RESET_MASK) && --timeout) {
		continue;
	}
	
	if (0 == timeout) {
		printk(KERN_ERR "%s: TIMEOUT\n", __FUNCTION__);
		return -ETIMEDOUT;
	}

    // Update USB PHY configuration
	// Change value of new register
	regVal = mv_raw_read_reg32(MV_USB_BRIDGE_IPG_REG(iDevice)); 

	// Change bits[14:8] - IPG for non Start of Frame Packets from 0x9(default) to 0xC 
	regVal &= ~(0x7F << 8);
	regVal |= (0xC << 8);

	mv_raw_write_reg32(MV_USB_BRIDGE_IPG_REG(iDevice), regVal);

    // USB 2.0 Power Control Register
	regVal= mv_raw_read_reg32(MV_USB_PHY_POWER_CTRL_REG(iDevice));
    // Bits 7:6 (BG_VSEL) = 0x1
	regVal &= ~(0x3 << 6);
	regVal |= (0x1 << 6);

	mv_raw_write_reg32(MV_USB_PHY_POWER_CTRL_REG(iDevice),regVal);

    // USB PHY Tx Control Register Register
    regVal = mv_raw_read_reg32(MV_USB_PHY_TX_CTRL_REG(iDevice)); 
	regVal |= (1 << 21);
	// Force Auto calibration to 0x8 */
    // Bit[13] = 0x1, (REG_EXT_RCAL_EN = 0x1)
	regVal |= (1<<13);
    // Bit[6:3] = 0x8 (IMP_CAL = 0x8)
	regVal &= ~(0xf << 3);
	regVal |= (0x8 << 3);

	mv_raw_write_reg32(MV_USB_PHY_TX_CTRL_REG(iDevice), regVal); 

    // USB PHY Rx Control Register
    regVal = mv_raw_read_reg32(MV_USB_PHY_RX_CTRL_REG(iDevice)); 
    // bit[8:9] - (DISCON_THRESHOLD ).
    // Orion1-A0/A1/B0 = 11, Orion2-A0 = 10, Orion1-B1 and Orion2-B0 later = 00
    regVal &= ~(0x3 << 8);
    // bit[21] = 0 (CDR_FASTLOCK_EN = 0).
    regVal &= ~(1 << 21);
    // bit[27:26] = 00 ( EDGE_DET_SEL = 00).
    regVal &= ~(0x3 << 26);
    // Bits[31:30] = RXDATA_BLOCK_LENGHT = 0x3
    regVal |= (0x3 << 30);
    // Bits 7:4 (SQ_THRESH) = 0x1
	regVal &= ~(0xf << 4);
	regVal |= (0x1 << 4);
	
    mv_raw_write_reg32(MV_USB_PHY_RX_CTRL_REG(iDevice), regVal);

    // USB PHY IVREF Control Register
    regVal = mv_raw_read_reg32(MV_USB_PHY_IVREF_CTRL_REG(iDevice)); 
   // Bits[1:0] = 0x2 (PLLVDD12 = 0x2)
    regVal &= ~(0x3 << 0);
    regVal |= (0x2 << 0);
	// Bits 5:4 (RXVDD) = 0x3
    regVal &= ~(0x3 << 4);
    regVal |= (0x3 << 4);
    // Bit 19 (Reserved) = 0x0
	regVal &= ~(0x1 << 19);
    
    mv_raw_write_reg32(MV_USB_PHY_IVREF_CTRL_REG(iDevice), regVal);

    // USB PHY TEST GROUP CONTROL Register
    regVal = mv_raw_read_reg32(MV_USB_PHY_TEST_GROUP_CTRL_REG_0(iDevice)); 
	// bit[15] = 0 (REG_FIFO_SQ_RST = 0).
    regVal &= ~(1 << 15);

    mv_raw_write_reg32(MV_USB_PHY_TEST_GROUP_CTRL_REG_0(iDevice), regVal);

    // Go through all windows in user table
	for(winNum=0;winNum<3;winNum++)
    {
		size = mv_raw_read_reg32( MV_USB_WIN_CTRL_REG(iDevice, winNum) );
		// The base address read should be the same as those coded (missing some bits though?)
		base = mv_raw_read_reg32( MV_USB_WIN_BASE_REG(iDevice, winNum) );
	
	    mv_raw_write_reg32( MV_USB_WIN_CTRL_REG(iDevice, winNum), size);
	    mv_raw_write_reg32( MV_USB_WIN_BASE_REG(iDevice, winNum), baseout[winNum]);
    }
  
    regVal = mv_raw_read_reg32(MV_USB_CORE_MODE_REG(iDevice)); 
    regVal &= ~MV_USB_CORE_MODE_MASK;
	
    // Set HOST mode
    regVal |= MV_USB_CORE_MODE_HOST; 

    // Set DEVICE mode not supported here
	// regVal |= MV_USB_CORE_MODE_DEVICE;

    // Workarround for USB underrun problem
    regVal |= MV_USB_CORE_STREAM_DISABLE_MASK;

    mv_raw_write_reg32(MV_USB_CORE_MODE_REG(iDevice), regVal); 
	
	return 0;
}

static void mv5182_setup(struct usb_hcd *hcd)
{
	struct ehci_hcd	*ehci = hcd_to_ehci(hcd);
	u32 __iomem	*reg;
	u32	tmp;

	reg = (u32 __iomem *)(((u8 __iomem *)ehci->regs) + 0x68);
	tmp = ehci_readl(ehci, reg);
    // Streaming disable for all USB rev0 owing to SOC issues
    tmp |= (1 << 4);
	ehci_writel(ehci, tmp, reg);
}

static int ehci_platform_probe(const struct hc_driver *driver, struct usb_hcd **hcd_out, struct platform_device *pdev)
{ 
    int result; 
    struct usb_hcd *hcd = NULL; 

	result = mv5182_start_ehc(pdev);
	if (result !=0)
		return result;
	
    hcd = usb_create_hcd (driver, &pdev->dev, pdev->dev.bus_id);
    if (hcd == NULL)  { 
        printk("%s: hcd_alloc failed\n", __FILE__); 
        return -ENOMEM; 
    } 
 
	hcd->irq = pdev->resource[1].start; 
	hcd->rsrc_start = pdev->resource[0].start;
	hcd->rsrc_len = pdev->resource[0].end - pdev->resource[0].start + 1;

	if (!request_mem_region(hcd->rsrc_start, hcd->rsrc_len, pdev->dev.bus_id)) {
		printk("%s: request_mem_region failed\n", __FILE__);
		result = -EBUSY;
		goto err1;
	}

	hcd->regs = ioremap(hcd->rsrc_start, hcd->rsrc_len);
	if (!hcd->regs) {
		printk("%s: ioremap failed\n", __FILE__);
		result = -ENOMEM;
		goto err2;
	}
	
    result = usb_add_hcd (hcd, hcd->irq, IRQF_DISABLED | IRQF_SHARED); //SA_INTERRUPT | SA_SHIRQ);
	if (result != 0) {
        printk("%s: usb_add_hcd failed, retval=0x%X\n", __FILE__, result); 
		goto fail_add_hcd;
    }   

	mv5182_setup(hcd);
	return result;
	
fail_add_hcd:
	iounmap(hcd->regs);
 err2:
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
 err1:
	usb_put_hcd(hcd);

    return result; 
} 
 
static int ehci_platform_remove(struct platform_device *pdev)
{ 
    struct usb_hcd *hcd = platform_get_drvdata(pdev); 
 
    usb_remove_hcd (hcd); 
	iounmap(hcd->regs);
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
    usb_put_hcd (hcd);
 
    return 0; 
} 

// Handle device setup and initialisation here as we are not a PCI device
static int ehci_setup(struct usb_hcd *hcd)
{
	struct ehci_hcd	*ehci = hcd_to_ehci(hcd);
	int	retval;

	ehci->caps = hcd->regs;
	ehci->regs = hcd->regs + HC_LENGTH(ehci_readl(ehci, &ehci->caps->hc_capbase));

	// cache this readonly data; minimize chip reads
	ehci->hcs_params = ehci_readl(ehci, &ehci->caps->hcs_params);

	retval = ehci_halt(ehci);
	if (retval)
		return retval;

	// data structure init
	retval = ehci_init(hcd);
	if (retval)
		return retval;

	ehci_reset(ehci);

	// Transaction Translator set
    ehci->is_tdi_rh_tt = 1; 

#ifdef	CONFIG_USB_SUSPEND
	if (ehci->no_selective_suspend && device_can_wakeup(&pdev->dev))
		ehci_warn(ehci, "selective suspend/wakeup unavailable\n");
#endif

	ehci_port_power(ehci, 0);
	
	return 0;
}

static const struct hc_driver ehci_mv5182_hc_driver = {
	.description		= hcd_name,
	.product_desc   	= "Marvell 5182 SOC EHCI Host Controller",
	.hcd_priv_size		= sizeof(struct ehci_hcd),
	.irq				= ehci_irq,
	.flags				= HCD_MEMORY | HCD_USB2,
	.reset				= ehci_setup,
	.start				= ehci_run,
	.stop				= ehci_stop,
	.shutdown			= ehci_shutdown,
	.urb_enqueue		= ehci_urb_enqueue,
	.urb_dequeue		= ehci_urb_dequeue,
	.endpoint_disable	= ehci_endpoint_disable,
	.get_frame_number	= ehci_get_frame,
	.hub_status_data	= ehci_hub_status_data,
	.hub_control		= ehci_hub_control,
#ifdef	CONFIG_PM
	.bus_suspend		= ehci_bus_suspend,
	.bus_resume			= ehci_bus_resume,
#endif
};

static int ehci_hcd_mv5182_start(struct platform_device *pdev)
{
	struct usb_hcd *hcd = NULL;
	
	if (usb_disabled())
		return -ENODEV;

	return (ehci_platform_probe(&ehci_mv5182_hc_driver, &hcd, pdev));
}

 /*TBD*/
/*static int ehci_hcd_mv5182_drv_suspend(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct usb_hcd *hcd = dev_get_drvdata(dev);

	return 0;
}
static int ehci_hcd_mv5182_drv_resume(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct usb_hcd *hcd = dev_get_drvdata(dev);

	return 0;
}
*/

MODULE_ALIAS("mv5182_ehci");

static struct platform_driver mv5182_ehci_driver =  
{
	.driver = {
	    .owner = THIS_MODULE,
		.name = "mv5182_ehci",
		.bus = &platform_bus_type
	},
    .probe = ehci_hcd_mv5182_start, 
    .remove = ehci_platform_remove, 
	.shutdown = usb_hcd_platform_shutdown,
#ifdef	CONFIG_PM
//	.bus_suspend = ehci_hcd_mv5182_drv_suspend,
//	.bus_resume = ehci_hcd_mv5182_drv_resume,
#endif
};  

